//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2018 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// 30th May 2018: Simulation of a practical massive MIMO system. Simulation
//                is limited by the memory (RAM) of the computer.
//                Frequency offset and channel is estimated.
// 11th Dec 2018: Preamble amplitude modified. Number of frequency bins added
//                to file name.
// Code tested on SCILAB ver 6.0.0.
//******************************************************************************
timer();
//******************************************************************************
// Add a word of caution.
//******************************************************************************
mprintf('Remember to comment out the SoF plots\n');
mprintf('Running this code will append to the data files\n');
mprintf('Is this ok?.\n');
mprintf('\n');
pause;
//******************************************************************************
// Initialize the random number generators with the date.
//******************************************************************************
n=getdate("s");
grand("setsd",n);
//******************************************************************************
// Constants.
//******************************************************************************
SNR_DB=4.0;                                               // SNR per bit in dB.
MAX_FRAMES=1000;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                    // Length of the preamble.
OFDM_SIZE=1024;           // Length of OFDM symbol, excluding the cyclic prefix.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=OFDM_SIZE+EST_CHAN_MEM;        // Frame size including cyclic prefix.
RX_DATA_SIZE=FRAME_SIZE+CHAN_LEN-1;                       // Length of rx. data.
MAX_SoF_PLOT_INX=10;                                            // Must be even.
NUM_TX=4;                                           // No. of transmit antennas.
NUM_RX=NUM_TX;                                       // No. of receive antennas.
NUM_RE_TX=2;                                          // No. of retransmissions.
PRE_AMB_QPSK_AMP=sqrt(4*PRE_AMB_SIZE/OFDM_SIZE);                  // Preamb amp.
//******************************************************************************
if NUM_TX > 8 then
 mprintf('No. of transmitters must not exceed 8.\n');
 quit;
elseif NUM_TX < 4 then
 mprintf('No. of transmitters must greater than or equal to 4.\n');
 quit;
end
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_MEM+CHAN_LEN-1;  // Length for freq. est.
MAX_FOFF=0.04;                                // Max. freq. offset search value.
MAX_FOFF1=0.03;                                            // Max. freq. offset.
ML_FOFF1_BINS=64;                                // No. of freq. bins in ML est.
//******************************************************************************
// Open files for appending.
//******************************************************************************
format(20);
x=string(SNR_DB);
y=strsubst(x, '.', '_');
y1=string(PRE_AMB_SIZE);
y2=string(NUM_TX);
y3=string(NUM_RE_TX);
y4=string(MAX_FRAMES);
y5=string(OFDM_SIZE);
y6=string(ML_FOFF1_BINS);
fd_u_ber_w =mopen(pwd()+'/u_ber_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_c_ber_w =mopen(pwd()+'/c_ber_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_f_eer_w =mopen(pwd()+'/f_eer_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_rf_w    =mopen(pwd()+'/rf_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_mf_w    =mopen(pwd()+'/mf_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_rf1_w   =mopen(pwd()+'/rf1_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_mf1_w   =mopen(pwd()+'/mf1_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
fd_ch_mse_w=mopen(pwd()+'/ch_mse_'+'SNR'+y+'_Lp'+y1+'_Nt'+y2+'_Re'+...
            y3+'_Fr'+y4+'_Ld'+y5+'_BINS'+y6+'.txt', 'at');
clear x y y1 y2 y3 y4 y5 y6;
//******************************************************************************
// SNR parameters.
//******************************************************************************
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
//******************************************************************************
// Variance per dimension of AWGN. Set noise variance and frequency offset
// equal to zero for testing.
//******************************************************************************
NOISE_VAR=2.0*NUM_TX*NUM_RE_TX*CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));
//NOISE_VAR=0.0;                                       // No noise. For testing.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
QPSK_Sym=repmat(matrix(repmat(QPSK_Map.',NUM_RX,1),1,CONSTELL_SIZE*NUM_RX),...
         OFDM_SIZE,1);
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
Pre_Amb_Sym=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym=Pre_Amb_Sym*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1,"nonsymmetric").'; // Division by PRE_AMB_SIZE.
P_Chan_Ip=[Pre_Amb_IFFT Pre_Amb_IFFT(1:EST_CHAN_MEM)];     // Preamb. chan. i/p.
//******************************************************************************
// Get the channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE,EST_CHAN_LEN);                  // Initialize.
Pre_Amb_IFFT_Suf=[Pre_Amb_IFFT Pre_Amb_IFFT(1:EST_CHAN_MEM)];
for i=1:PRE_AMB_SIZE
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT_Suf(EST_CHAN_LEN+i-1:-1:i);
end
clear Pre_Amb_IFFT_Suf;
//******************************************************************************
// Least squares channel estimation matrix for 1st transmitter.
//******************************************************************************
Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// Get the matched filter for the preamble.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
MF_Coeff=MF_Coeff/MF_Energy;                       // Normalized to unit energy.
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x y;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
RMS_Err_FOFF=0.0;                       // RMS individual freq offset est error.
Max_Err_FOFF= -1.0;                     // Max individual freq offset est error.
RMS_Err_FOFF1=0.0;                             // RMS avg freq offset est error.
Max_Err_FOFF1= -1.0;                           // Max avg freq offset est error.
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D^2)/(1+D+D^2)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 0 1];                                           // Numerator coefficients.
DC=[1 1 1];                                         // Denominator coefficients.
NC_Size=size(NC);
NUM_INPUTS=2;                   // No. of input bit combinations in the trellis.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op,Ga_Inx,Gb_Inx]=Get_Trellis();
RSC_Gen_Vec=Get_RSC_Gen_Vec();                          // Get impulse response.
//******************************************************************************
// Get the interleaver and deinterleaver map for the turbo code.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(OFDM_SIZE);
//******************************************************************************
// Get the data interleaver and deinterleaver map for OFDM.
//******************************************************************************
[OFDM_Iv_Map,OFDM_Div_Map]=Get_OFDM_Iv_Div_Map(OFDM_SIZE);
//******************************************************************************
// Cyclic prefix (CP) indices.
//******************************************************************************
CP_ST_INX=OFDM_SIZE-EST_CHAN_MEM+1;         // Start index for extraction of CP.
CP_ED_INX=OFDM_SIZE;                          // End index for extraction of CP.
//******************************************************************************
// Channel estimation error.
//******************************************************************************
Chan_Est_Err=0.0;
//******************************************************************************
// Channel output matrix size.
//******************************************************************************
CHAN_OP_SIZE=FOFF_EST_RX_SIZE*NUM_RE_TX*NUM_TX*NUM_RX;
//******************************************************************************
// Minimum exponent.
//******************************************************************************
MIN_EXPO= -30.0;
//******************************************************************************
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Erasure=0;                                          // Number of frame erasures.
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
//******************************************************************************
// Get the random frequency offset and the frequency offset vector.
//******************************************************************************
 FOFF=2.0*MAX_FOFF1*(grand(1,1,'def')-0.5);  // FOFF in [-MAX_FOFF1, MAX_FOFF1].
// FOFF=0.0;                                       // For SoF plots and testing.
 FOFF_Vec=0:FOFF_EST_RX_SIZE-1;
 FOFF_Vec=FOFF_Vec*%i*FOFF;
 FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
//******************************************************************************
// Estimate the channel and frequency offsets.
//******************************************************************************
 Chan=Get_Channel();
 Chan_Op=[Chan;zeros(FOFF_EST_RX_SIZE-CHAN_LEN,NUM_TX,NUM_RX,NUM_RE_TX)];
 Chan_Op1=resize_matrix(convol(P_Chan_Ip,Chan_Op(:)),...
          1,CHAN_OP_SIZE).* ...
          repmat(FOFF_Vec,1,NUM_TX*NUM_RX*NUM_RE_TX) + ...
          grand(1,CHAN_OP_SIZE,'nor',NOISE_MEAN,NOISE_STD) + %i* ...
          grand(1,CHAN_OP_SIZE,'nor',NOISE_MEAN,NOISE_STD);
 clear Chan_Op;
 Chan_Op2=repmat(matrix(Chan_Op1,FOFF_EST_RX_SIZE,NUM_TX,NUM_RX,NUM_RE_TX),...
          ML_FOFF1_BINS+1,1,1,1).* ...
          repmat(FOFF_Est_Matrix.'(:),1,NUM_TX,NUM_RX,NUM_RE_TX);
 FOFF_Sync_Op=matrix(resize_matrix(convol(MF_Coeff,Chan_Op2(:)),1,...
              CHAN_OP_SIZE*(ML_FOFF1_BINS+1)),...
              FOFF_EST_RX_SIZE*(ML_FOFF1_BINS+1),NUM_TX*NUM_RX*NUM_RE_TX);
 clear Chan_Op2;
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op),'r');
 Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
 Est_FOFF=Est_FOFF_Inx*Freq_Step-MAX_FOFF;
// Get_SoF_Plot();                         // Set FOFF=0 to get a centered plot.
//******************************************************************************
// Average frequency offset estimate.
//******************************************************************************
 Avg_Est_FOFF1=sum(Est_FOFF)/(NUM_TX*NUM_RX*NUM_RE_TX);
// Avg_Est_FOFF1=0.0;                              // For SoF plots and testing.
 Avg_Freq_Est_Err1=FOFF-Avg_Est_FOFF1;
 RMS_Err_FOFF1=RMS_Err_FOFF1+Avg_Freq_Est_Err1^2;
 Max_Err_FOFF1=max([Max_Err_FOFF1 abs(Avg_Freq_Est_Err1)]);
 Avg_Est_FOFF_Vec=exp(-%i*[0:FOFF_EST_RX_SIZE-1]*Avg_Est_FOFF1);
//******************************************************************************
// Individual frequency offset estimates.
//******************************************************************************
 Freq_Est_Err=FOFF-Est_FOFF;
 RMS_Err_FOFF=RMS_Err_FOFF+sum(Freq_Est_Err.^2);
 Max_Err_FOFF=max([Max_Err_FOFF abs(Freq_Est_Err)]);
//******************************************************************************
// Timing estimates.
//******************************************************************************
 Diff_Inx=Max_Inx-PRE_AMB_SIZE-Est_FOFF_Inx*FOFF_EST_RX_SIZE;
 Erasure_Find=isempty(find((Diff_Inx < 0) | (Diff_Inx > CHAN_MEM)));
//******************************************************************************
 clear FOFF_Sync_Op Max_Sync_Op Max_Inx;
//******************************************************************************
// Get starting and ending indices for channel estimation. Do channel
// estimation.
//******************************************************************************
 if Erasure_Find == %T then                     // Empty matrix. Not an erasure.
  Start_Inx=CHAN_LEN+CHAN_MEM;
  End_Inx=Start_Inx+PRE_AMB_SIZE-1;
  Chan_Op3=matrix(Chan_Op1.*repmat(Avg_Est_FOFF_Vec,...
           1,NUM_TX*NUM_RX*NUM_RE_TX),FOFF_EST_RX_SIZE,NUM_TX*NUM_RX*NUM_RE_TX);
  clear Chan_Op1;
  Chan_Op4=Chan_Op3(Start_Inx:End_Inx,:);
  clear Chan_Op3;
  Chan_Est=matrix(Chan_Est_Matrix*Chan_Op4,EST_CHAN_LEN,...
           NUM_TX,NUM_RX,NUM_RE_TX);
  Est_Chan_FFT=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,...
               NUM_TX,NUM_RX,NUM_RE_TX)];
  Est_Chan_FFT=fft(Est_Chan_FFT,-1,1,"nonsymmetric");  // FFT along the 1st dim.
  Gains=squeeze(sum(sum(abs(Est_Chan_FFT).^2,3),4))/NUM_RE_TX;
  Est_C_Var=sum(abs(Est_Chan_FFT).^2)/size(Est_Chan_FFT,'*');    // 2D fade var.
  Chan_FFT=[Chan;zeros(OFDM_SIZE-CHAN_LEN,NUM_TX,NUM_RX,NUM_RE_TX)];
  Chan_FFT=fft(Chan_FFT,-1,1,"nonsymmetric");         // FFT along the 1st dim.
//  Plot_Channel();
  Chan_Est_Err=Chan_Est_Err + ...
               sum(abs([Chan; ...
               zeros(EST_CHAN_LEN-CHAN_LEN,NUM_TX,NUM_RX,NUM_RE_TX)] - ...
               Chan_Est).^2);
//******************************************************************************
// Compute noise variance in two dimensions, in the frequency domain.
//******************************************************************************
  Est_N_Var=OFDM_SIZE*sum(abs(Chan_Op4-Chan_Ip_Matrix*matrix(Chan_Est,...
            EST_CHAN_LEN,NUM_TX*NUM_RX*NUM_RE_TX)).^2)/size(Chan_Op4,'*');
  clear Chan_Op4;
//******************************************************************************
// Compute the noise+interference variance.
//******************************************************************************
  Eff_N_Var=(2*NUM_TX*(NUM_TX-1)*Est_C_Var^2+...
            NUM_TX*Est_C_Var*Est_N_Var)/NUM_RE_TX;
//******************************************************************************
// Generate coded QPSK symbols
//******************************************************************************
  [My_Code,My_Data]=Get_Coded_QPSK();
//******************************************************************************
// Get the channel output.
//******************************************************************************
  Chan_Op5=[Chan;zeros(OFDM_SIZE-CHAN_LEN,NUM_TX,NUM_RX,NUM_RE_TX)];
  Chan_Op6=squeeze(fft(fft(sum(fft(Chan_Op5,-1,1,"nonsymmetric").* ...
           repmat(My_Code,1,1,NUM_RX,NUM_RE_TX),2),1,1,"nonsymmetric").* ...
           repmat(exp(-%i*([0:OFDM_SIZE-1].')*Avg_Freq_Est_Err1),...
           1,1,NUM_RX,NUM_RE_TX) + ...
           grand(OFDM_SIZE,1,NUM_RX,NUM_RE_TX,'nor',NOISE_MEAN,NOISE_STD) + ...
           %i* ...
           grand(OFDM_SIZE,1,NUM_RX,NUM_RE_TX,'nor',NOISE_MEAN,NOISE_STD),...
           -1,1,"nonsymmetric"));
  clear Chan_Op5;
  Rx_Sig=squeeze(sum(squeeze(sum(matrix(repmat(Chan_Op6,1,NUM_TX,1),...
         OFDM_SIZE,NUM_TX,NUM_RX,NUM_RE_TX).* ...
         conj(permute(Est_Chan_FFT,[1 3 2 4])),2)),3))/NUM_RE_TX;
  clear Chan_Op6;
//******************************************************************************
// Compute squared Euclidean distances.
//******************************************************************************
  Dist=matrix(abs(repmat(Rx_Sig,1,CONSTELL_SIZE)- ...
       repmat(Gains,1,CONSTELL_SIZE).*QPSK_Sym).^2,...
       OFDM_SIZE,NUM_RX,CONSTELL_SIZE);
  clear Rx_Sig;
//******************************************************************************
// Compute the minimum over the 3rd dimension.
//******************************************************************************
  [M1,K1]=min(Dist(:,:,1),Dist(:,:,2),Dist(:,:,3),Dist(:,:,4));
//******************************************************************************
  Rx_Sym=real(matrix(QPSK_Map(K1(:,1:NUM_TX/2)),OFDM_SIZE,NUM_TX/2));
  Ber=Ber+nnz(My_Data(:,1:NUM_TX/2)-Rx_Sym(OFDM_Iv_Map(:),:));   // Uncoded BER.
  clear M1 K1;
//******************************************************************************
// Compute gammas.
//******************************************************************************
  My_Exp1= -Dist(OFDM_Iv_Map(:),1:NUM_RX/2,:)/Eff_N_Var;
  My_Exp2= -Dist(OFDM_Iv_Map(:),NUM_RX/2+1:$,:)/Eff_N_Var;
  My_Exp1=My_Exp1-repmat(max(My_Exp1(:,:,1),My_Exp1(:,:,2),...
          My_Exp1(:,:,3),My_Exp1(:,:,4)),1,1,CONSTELL_SIZE);
  My_Exp2=My_Exp2-repmat(max(My_Exp2(:,:,1),My_Exp2(:,:,2),...
          My_Exp2(:,:,3),My_Exp2(:,:,4)),1,1,CONSTELL_SIZE);
  Exp1_Inx=find(My_Exp1 < MIN_EXPO);
  Exp2_Inx=find(My_Exp2 < MIN_EXPO);
  My_Exp1(Exp1_Inx)= MIN_EXPO;
  My_Exp2(Exp2_Inx)= MIN_EXPO;
  My_Gamma1=exp(My_Exp1);
  My_Gamma2=exp(My_Exp2);
  clear My_Exp1 My_Exp2 Dist;
//******************************************************************************
// The MAP (BCJR) algorithm.
//******************************************************************************
  Prob=0.5*ones(OFDM_SIZE,NUM_RX/2,NUM_INPUTS);
//******************************************************************************
//******************************************************************************
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process_End(Prob);
//******************************************************************************
  [M2,K2]=max(Prob(:,:,1),Prob(:,:,2));               // Max. prob. and indices.
  Rx_Sym1=1-2*(K2-1);
  C_Ber=C_Ber+nnz(My_Data(:,1:NUM_RX/2)-Rx_Sym1);
  clear M2 K2 Prob My_Data Rx_Sym1 My_Gamma1 My_Gamma2;
//******************************************************************************
 else
  Erasure=Erasure+1;
 end
end
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF=sqrt(RMS_Err_FOFF/(MAX_FRAMES*NUM_TX*NUM_RX*NUM_RE_TX));
F_U_Ber=2*Ber/((MAX_FRAMES-Erasure)*OFDM_SIZE*NUM_TX);           // Uncoded BER.
F_C_Ber=2*C_Ber/((MAX_FRAMES-Erasure)*OFDM_SIZE*NUM_TX);           // Coded BER.
F_F_Eer=Erasure/MAX_FRAMES;                         // Frame erasure error rate.
Chan_Est_Err=Chan_Est_Err/((MAX_FRAMES-Erasure)*NUM_TX*NUM_RX* ...
             NUM_RE_TX*CHAN_LEN);
mfprintf(fd_u_ber_w,'%f %.20f\n',SNR_DB,F_U_Ber);
mfprintf(fd_c_ber_w,'%f %.20f\n',SNR_DB,F_C_Ber);
mfprintf(fd_f_eer_w,'%f %.20f\n',SNR_DB,F_F_Eer);
mfprintf(fd_rf1_w,'%f %.20f\n',SNR_DB,RMS_Err_FOFF1);
mfprintf(fd_rf_w,'%f %.20f\n',SNR_DB,RMS_Err_FOFF);
mfprintf(fd_mf1_w,'%f %.20f\n',SNR_DB,Max_Err_FOFF1);
mfprintf(fd_mf_w,'%f %.20f\n',SNR_DB,Max_Err_FOFF);
mfprintf(fd_ch_mse_w,'%f %.20f\n',SNR_DB,Chan_Est_Err);
mclose(fd_u_ber_w);
mclose(fd_c_ber_w);
mclose(fd_f_eer_w);
mclose(fd_rf1_w);
mclose(fd_rf_w);
mclose(fd_mf1_w);
mclose(fd_mf_w);
mclose(fd_ch_mse_w);
timer()
//******************************************************************************
